-module(tpTermCut).

%% 现在要使用硬编码的值。我们不能花时间在记录或地图中查找。

-export([cut/1]).
-export([cut/2]).



cut(Term) ->
   cut(Term, 1).

cut(_, Depth) when Depth > ?MaxDepth ->
   '$truncated11';
cut(Bits, _) when is_bitstring(Bits), bit_size(Bits) > ?MaxBitSize ->
   <<CutBits:?MaxBinSize/binary, _/bits>> = Bits,
   <<CutBits/binary, "$truncated">>;
cut(Term, Depth) when is_list(Term), Depth =:= ?MaxDepth ->
   ['$truncated'];
cut(Term, Depth) when is_list(Term) ->
   cutList(Term, Depth, 0, ?MaxListSize, 0, []);
cut(Term, Depth) when is_map(Term), Depth =:= ?MaxDepth ->
   #{'$truncated' => '$truncated'};
cut(Term, Depth) when is_map(Term) ->
   maps:from_list(cutMap(maps_to_list(Term, ?MaxMapSize), Depth, 0));
cut(Term, Depth) when is_tuple(Term), Depth =:= ?MaxDepth ->
   {'$truncated'};
cut(Term, Depth) when is_tuple(Term) ->
   list_to_tuple(cutList(tuple_to_list(Term), Depth, 0, ?MaxTupleSize, 0, []));
cut(Term, _) ->
   Term.

cutList([], _, _, _, _, Acc) ->
   lists:reverse(Acc);
cutList([Term | Tail], Depth, Len, MaxLen, NumStructs, Acc) ->
   case Len > MaxLen orelse NumStructs > ?MaxNestStruct of
      true ->
         lists:reverse(['$truncated' | Acc]);
      _ ->
         %% if List was a cons, Tail can be anything
         cutList(Tail, Depth, Len + 1, MaxLen, NumStructs + isStruct(Term), [cut(Term, Depth + 1) | Acc])
   end;
cutList(Term, Depth, _, _, _, Acc) -> %% if List was a cons  [[[a,a,a|b],1,2,3]|bbbb]
   lists:reverse(Acc) ++ cut(Term, Depth + 1).

cutMap([], _, _) ->
   [];
cutMap(_, _, NumStructs) when NumStructs > ?MaxNestStruct ->
   [{'$truncated', '$truncated'}];
cutMap([{Key, Value} | Tail], Depth, NumStructs) ->
   AddStruct = isStruct(Key) + isStruct(Value),
   [{cut(Key, Depth + 1), cut(Value, Depth + 1)} | cutMap(Tail, Depth, NumStructs + AddStruct)].

isStruct(Term) ->
   case is_list(Term) orelse is_map(Term) orelse is_tuple(Term) of
      true ->
         1;
      _ ->
         0
   end.

maps_to_list(Map, MaxSize) ->
   I = maps:iterator(Map),
   maps_to_list(maps:next(I), MaxSize, []).

%% Returns elements in arbitrary order. We reverse when we truncate
%% so that the truncated elements come at the end to avoid having
%% two truncated elements in the final output.
maps_to_list(none, _, Acc) ->
   Acc;
maps_to_list(_, 0, Acc) ->
   lists:reverse([{'$truncated', '$truncated'} | Acc]);
maps_to_list({K, V, I}, N, Acc) ->
   maps_to_list(maps:next(I), N - 1, [{K, V} | Acc]).
